// Copyright (c) 2016 University of Helsinki
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 3 of the License, or (at your option) any later version.
// See the file COPYING included with this distribution for more
// information.

#include "HfstDataTypes.h"
#include <iosfwd>

namespace hfst
{
  void print_dot(FILE* out, HfstTransducer& t);
  void print_dot(std::ostream & oss, HfstTransducer& t);
}
