// Copyright (c) 2016 University of Helsinki
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 3 of the License, or (at your option) any later version.
// See the file COPYING included with this distribution for more
// information.

/*
 * This is a convenience header file for including common HFST headers.
 */

#ifndef _HFST_H
#define _HFST_H

#include "HfstTransducer.h"
#include "HfstInputStream.h"
#include "HfstOutputStream.h"
#include "HfstSymbolDefs.h"
#include "HfstExceptionDefs.h"

#endif // HFST_H
